﻿
/*****************************************************************************

This class has been written by Elmü (elmue@gmx.de)

Check if you have the latest version on:
https://www.codeproject.com/Articles/5293980/Graph3D-A-Windows-Forms-Render-Control-in-Csharp

=======================================
 
IMPORTANT:
This class has been written by a software developer with 40 years programming exprience.
This class is optimized for the highest possible speed in every line of it's code.
If you found this code on Github or else where you do NOT have the original high quality code!
This class with 2000 lines is extremely complex and there is a very high risk that a beginner has 
broken this code by modifying it without properly understanding it.

=======================================
 
NAMING CONVENTIONS which allow to see the variable type at a glance:
 
     cName  for class    definitions
     tName  for type     definitions
     eName  for enum     definitions
     kName  for struct   definitions
   delName  for delegate definitions

    b_Name  for bool
    c_Name  for Char, also Color
    d_Name  for double
    e_Name  for enum variables
    f_Name  for function delegates
    i_Name  for instances of classes
    k_Name  for "konstructs" (struct)
	r_Name  for Rectangle
    s_Name  for strings
 
   s8_Name  for   signed  8 Bit (SByte)
  s16_Name  for   signed 16 Bit (short)
  s32_Name  for   signed 32 Bit (int)
  s64_Name  for   signed 64 Bit (long)
   u8_Name  for unsigned  8 Bit (Byte)
  u16_Name  for unsigned 16 bit (UInt16)
  u32_Name  for unsigned 32 Bit (uint)
  u64_Name  for unsigned 64 Bit (ulong)

  An additional "m" is prefixed for all member variables (e.g. ms_String)


*****************************************************************************/

using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Text;
using System.Globalization;
using System.Windows.Forms;
using System.Diagnostics;

namespace Plot3D
{
    /// <summary>
    /// ATTENTION: This class is not thread safe.
    /// Call all functions only from the GUI thread or use Control.Invoke()
    /// </summary>
    public class Graph3D : UserControl
    {
        #region enums

        public enum eRaster
        {
            Off,      // draw no coordinate system
            MainAxis, // draw solid axis X,Y,Z
            Raster,   // draw axis and dotted raster lines
            Labels,   // draw axis and dotted raster lines and labels in quadrant 3
        }

        /// <summary>
        /// If a function has an asymetric range for X and Y as example "Callback" a separate normalization 
        /// would always lead to a square X,Y pane which would be a distortion for the relation between X and Y values.
        /// MaintainXY  guarantees that the relation between X and Y values is maintained.
        /// MaintainXYZ additionally guarantees that the relation between X, Y and Z values is maintained.
        /// </summary>
        public enum eNormalize
        {
            Separate,    // Normalize X,Y,Z separately (use this for discrete values)
            MaintainXY,  // Normalize X,Y   without changing their relation (use this for functions)
            MaintainXYZ, // Normalize X,Y,Z without changing their relation (use this for functions)
        }

        /// <summary>
        /// This is for drawing surface plots and formulas
        /// </summary>
        public enum eSurfaceMode
        {
            // 0 is invalid
            Fill  = 1, // Fill surface polygons with color scheme and draw lines with mi_LinePen (default = black)
            Lines = 2, // Draw only lines with color scheme 
        }

        /// <summary>
        /// This is for drawing scatter plots
        /// </summary>
        public enum eScatterMode
        {
            // 0 is invalid
            Shapes = 1, // Each scatter point is drawn as circle, square or triangle as defined with eScatterShape
            Lines  = 2, // neighboured scatter points are drawn as connected line
        }

        /// <summary>
        /// The shape of scatter points
        /// </summary>
        public enum eScatterShape
        {
            // 0 is invalid
            Circle   = 1,
            Square   = 2,
            Triangle = 3,
        }

        public enum eCoord
        {
            // for axis in coordinate system
            X = 0,
            Y = 1,
            Z = 2,
            Invalid,
        }

        enum eMouseAction
        {
            None = 0,
            Move,
            Rho,
            Theta,
            Phi,
        }

        #endregion

        // ================== PUBLIC =================

        #region cPoint3D

        public class cPoint3D
        {
            public double md_X;
            public double md_Y;
            public double md_Z;

            public cPoint3D()
            {
            }

            public cPoint3D(double X, double Y, double Z)
            {
                md_X = X;
                md_Y = Y;
                md_Z = Z;
            }

            public cPoint3D Clone()
            {
                return new cPoint3D(md_X, md_Y, md_Z);
            }

            public bool Equals(cPoint3D i_Point)
            {
                return md_X == i_Point.md_X && md_Y == i_Point.md_Y && md_Z == i_Point.md_Z;
            }

            public double GetValue(eCoord e_Coord)
            {
                switch (e_Coord)
                {
                    case eCoord.X: return md_X;
                    case eCoord.Y: return md_Y;
                    case eCoord.Z: return md_Z;
                    default:       return 0;
                }
            }
            public void SetValue(eCoord e_Coord, double d_Value)
            {
                switch (e_Coord)
                {
                    case eCoord.X: md_X = d_Value; break;
                    case eCoord.Y: md_Y = d_Value; break;
                    case eCoord.Z: md_Z = d_Value; break;
                }
            }

            // For debugging in Visual Studio
            public override string ToString()
            {
                return String.Format("{0}, {1}, {2}", FormatDouble(md_X), FormatDouble(md_Y), FormatDouble(md_Z));
            }
        }

        #endregion

        #region cColorScheme

        /// <summary>
        /// Pens and Brushes are GDI objects which must NOT be created on the fly in Draw() !
        /// For speed optimization these are created once only.
        /// </summary>
        public class cColorScheme
        {
            private Brush[] mi_Brushes;
            private Pen[]   mi_Pens;

            /// <summary>
            /// Constructor
            /// If you want to draw the entire graph with only one color, you can pass a single Color here.
            /// </summary>
            public cColorScheme(int s32_LineWidth, params Color[] c_Colors)
            {
                if (s32_LineWidth <= 0 || c_Colors.Length == 0)
                    throw new ArgumentException();

                mi_Brushes = new SolidBrush[c_Colors.Length];
                mi_Pens    = new Pen       [c_Colors.Length];

                for (int i=0; i < c_Colors.Length; i++)
                {
                    mi_Brushes[i] = new SolidBrush(c_Colors[i]);
                    mi_Pens   [i] = new Pen       (c_Colors[i], s32_LineWidth); 
                }
            }

            public Brush GetBrush(int s32_Index)
            {
                if (s32_Index < 0) return Brushes.Goldenrod;
                else               return mi_Brushes[s32_Index];
            }

            public Pen GetPen(int s32_Index)
            {
                if (s32_Index < 0) return Pens.Goldenrod;
                else               return mi_Pens[s32_Index];
            }

            public int CalcIndex(double d_FactorZ)
            {
                if (double.IsNaN(d_FactorZ))
                    return -1;

                d_FactorZ = Math.Min(1.0, d_FactorZ);
                d_FactorZ = Math.Max(0.0, d_FactorZ);

                // d_FactorZ is a value between 0.0 and 1.0
                return (int)(d_FactorZ * (mi_Brushes.Length - 1));
            }
        }

        #endregion

        // ----- 3D Input Data-------

        #region cRenderData

        /// <summary>
        /// Base class for cSurfaceData, cScatterData, cLineData, cPolygonData
        /// </summary>
        public class cRenderData
        {
            public Graph3D mi_Inst;

            // 1.) must be called first
            public virtual void CalcMinMax()
            {
                throw new NotImplementedException();
            }

            // 2.) When this is called mi_MinMax has valid values
            public virtual void PreProcess()
            {
                throw new NotImplementedException();
            }

            // 3.) called from Render(Graphics)
            public virtual void ProjectDrawObjects()
            {
                throw new NotImplementedException();
            }
        }

        #endregion

        #region cSurfaceData

        public class cSurfaceData : cRenderData
        {
            private eSurfaceMode   me_Mode;
            private int            ms32_Cols;
            private int            ms32_Rows;
            private cColorScheme   mi_ColorScheme;
            private cPoint[,]      mi_PointArray;
            private cPolygon[]     mi_Polygons;
            private Pen            mi_BorderPen = new Pen(Color.Black, 1);  

            // This Pen draws the thin black lines around the border of the filled polygons.
            // You can set the Pen to null and no black lines will be drawn.
            public Pen BorderPen
            {
                set { mi_BorderPen = value; }
            }
            public int Cols
            {
                get { return ms32_Cols; }
            }
            public int Rows
            {
                get { return ms32_Rows; }
            }

            public cSurfaceData(eSurfaceMode e_Mode, int s32_Cols, int s32_Rows, cColorScheme i_ColorScheme)
            {
                if (s32_Cols < 2 || s32_Rows < 2 || i_ColorScheme == null)
                    throw new ArgumentException();

                me_Mode        = e_Mode;
                ms32_Cols      = s32_Cols;
                ms32_Rows      = s32_Rows;
                mi_ColorScheme = i_ColorScheme;
                mi_PointArray  = new cPoint[s32_Cols, s32_Rows];
                mi_Polygons    = new cPolygon[(ms32_Cols - 1) * (ms32_Rows - 1)];
            }

            public void SetPointAt(int s32_Column, int s32_Row,        // Position in the array
                                   double d_X, double d_Y, double d_Z, // Values of the 3D Point
                                   String s_UserTip = "")              // optional additional user defined tooltip
            {
                mi_PointArray[s32_Column, s32_Row] = new cPoint(d_X, d_Y, d_Z, 1, s_UserTip);
            }

            public override void CalcMinMax()
            {
                for (int C = 0; C < ms32_Cols; C++)
                {
                    for (int R = 0; R < ms32_Rows; R++)
                    {
                        cPoint i_Point = mi_PointArray[C,R];
                        if (i_Point == null)
                            throw new Exception("Surface point at column " + C + ", row " + R + " not set.");

                        mi_Inst.mi_MinMax.AddPoint(i_Point.mi_P3D);
                    }
                }
            }

            public override void PreProcess()
            {
                mi_Inst.mi_MinMax.ms32_TotPolygons += mi_Polygons.Length;

                foreach (cPoint i_Point in mi_PointArray)
                {
                    mi_Inst.mi_Tooltip.AddPoint(i_Point);
                }

                int P = 0;
                for (int C = 0; C < ms32_Cols -1; C++)
                {
                    for (int R = 0; R < ms32_Rows -1; R++)
                    {
                        cPolygon i_Poly = new cPolygon(me_Mode, mi_PointArray[C,   R],
                                                                mi_PointArray[C,   R+1],
                                                                mi_PointArray[C+1, R+1],
                                                                mi_PointArray[C+1, R]);
                        mi_Polygons[P++] = i_Poly;

                        cPolygon2D i_Poly2D = i_Poly.mi_Polygon2D;
                        double  d_FactorZ   = mi_Inst.mi_MinMax.CalcFactorZ(i_Poly2D.md_AvrgZ);
                        int     s32_Index   = mi_ColorScheme.CalcIndex(d_FactorZ);
                        i_Poly2D.mi_Brush   = mi_ColorScheme.GetBrush (s32_Index);

                        if (me_Mode == eSurfaceMode.Fill)
                            i_Poly2D.mi_Pen = mi_BorderPen; // thin black
                        else
                            i_Poly2D.mi_Pen = mi_ColorScheme.GetPen(s32_Index); // user defined
                    }
                }
            }

            public override void ProjectDrawObjects()
            {
                // Project polygons
                foreach (cPolygon i_Poly in mi_Polygons)
                {
                    i_Poly.Project3D(mi_Inst);

                    i_Poly.mi_Polygon2D.CalcSortOrder(mi_Inst);

                    mi_Inst.InsertDrawObject(i_Poly.mi_Polygon2D);
                }
            }
        }

        #endregion

        #region cScatterData

        public class cScatterData : cRenderData
        {
            private List<cScatter> mi_Points = new List<cScatter>();
            private eScatterMode   me_Mode;
            private cColorScheme   mi_ColorScheme;

            /// <summary>
            /// If all Scatter points contain a valid Brush or Pen, you can pass i_ColorScheme == null here
            /// </summary>
            public cScatterData(eScatterMode e_Mode, cColorScheme i_ColorScheme)
            {
                me_Mode        = e_Mode;
                mi_ColorScheme = i_ColorScheme;
            }

            /// <summary>
            /// s_UserTip is an optional additional user defined tooltip
            /// </summary>
            public void AddShape(double X, double Y, double Z, eScatterShape e_Shape, int s32_Radius, Brush i_Brush, String s_UserTip = "")
            {
                if (me_Mode != eScatterMode.Shapes) 
                    throw new Exception("Invalid function called for the selected ScatterMode.");

                if (i_Brush == null && mi_ColorScheme == null)
                    throw new Exception("You must specify a Brush or a ColorScheme");

                mi_Points.Add(new cScatter(X, Y, Z, e_Shape, s32_Radius, i_Brush, s_UserTip));
            }

            /// <summary>
            /// b_Append = true  --> append to list
            /// b_Append = false --> insert at begin of list
            /// ATTENTION: The order matters in which the line points are added!
            /// s_UserTip is an optional additional user defined tooltip
            /// </summary>
            public void AddLine(double X, double Y, double Z, bool b_Append, Pen i_Pen, String s_UserTip = "")
            {
                if (me_Mode != eScatterMode.Lines) 
                    throw new Exception("Invalid function called for the selected ScatterMode.");

                if (i_Pen == null && mi_ColorScheme == null)
                    throw new Exception("You must specify a Pen or a ColorScheme");

                float f_Width;
                if (i_Pen != null) f_Width = i_Pen.Width;
                else               f_Width = mi_ColorScheme.GetPen(0).Width;

                cScatter i_Scatter = new cScatter(X, Y, Z, i_Pen, (int)f_Width / 2, s_UserTip);

                if (b_Append) mi_Points.Add(i_Scatter);
                else          mi_Points.Insert(0, i_Scatter);
            }

            public override void CalcMinMax()
            {
                foreach (cScatter i_Scatter in mi_Points)
                {
                    mi_Inst.mi_MinMax.AddPoint(i_Scatter.mi_Point.mi_P3D);
                }
            }

            public override void PreProcess()
            {
                foreach (cScatter i_Scatter in mi_Points)
                {
                    // Load missing Pens and Brushes from ColorScheme which the user has not specified.
                    if (mi_ColorScheme != null)
                    {
                        double d_FactorZ = mi_Inst.mi_MinMax.CalcFactorZ(i_Scatter.md_AvrgZ);
                        int    s32_Index = mi_ColorScheme.CalcIndex(d_FactorZ);

                        if (i_Scatter.mi_Pen   == null) i_Scatter.mi_Pen   = mi_ColorScheme.GetPen  (s32_Index);
                        if (i_Scatter.mi_Brush == null) i_Scatter.mi_Brush = mi_ColorScheme.GetBrush(s32_Index);
                    }

                    mi_Inst.mi_Tooltip.AddPoint(i_Scatter.mi_Point);
                }

                if (me_Mode == eScatterMode.Lines)
                {
                    // Build a single chained list of all lines in the original order in which they have been added.
                    cScatter i_Prev = null;
                    foreach (cScatter i_Scatter in mi_Points)
                    {
                        i_Scatter.mi_Previous = i_Prev; 
                        i_Prev = i_Scatter;

                        i_Scatter.mi_Pen.StartCap = LineCap.Round;
                        i_Scatter.mi_Pen.EndCap   = LineCap.Round;
                    }

                    mi_Inst.mi_MinMax.ms32_TotLines  += mi_Points.Count;
                }
                else
                {
                    mi_Inst.mi_MinMax.ms32_TotPoints += mi_Points.Count;
                }
            }

            public override void ProjectDrawObjects()
            {
                // Convert 3D --> 2D
                foreach (cScatter i_Scatter in mi_Points)
                {
                    i_Scatter.mi_Point.Project3D(mi_Inst);
                    i_Scatter.SetPoint2D(i_Scatter.mi_Point.mi_P2D);

                    i_Scatter.CalcSortOrder(mi_Inst);

                    mi_Inst.InsertDrawObject(i_Scatter);
                }
            }
        }

        #endregion

        #region cLineData

        public class cLineData : cRenderData
        {
            private List<cLine>   mi_Lines  = new List<cLine>();
            private cColorScheme  mi_ColorScheme;
            private int           ms32_Lines;

            /// <summary>
            /// If you use only solid lines and specify a valid Pen, you can pass i_ColorScheme == null here
            /// </summary>
            public cLineData(cColorScheme i_ColorScheme)
            {
                mi_ColorScheme = i_ColorScheme;
            }

            /// <summary>
            /// Add a line which will be drawn entirely in one color
            /// </summary>
            public void AddSolidLine(cPoint3D i_Start, cPoint3D i_End, Pen i_Pen)
            {
                if (i_Pen == null && mi_ColorScheme == null)
                    throw new Exception("You must specify a Pen or a ColorScheme");

                cLine i_Line = new cLine(new cPoint(i_Start, 1, ""), 
                                         new cPoint(i_End,   1, ""), i_Pen);
                mi_Lines.Add(i_Line);
                ms32_Lines ++;
            }

            /// <summary>
            /// Add a line which will appear with multiple colors of the ColorScheme by splitting it into multiple parts.
            /// If s32_Parts = 50, the line is rendered in 50 parts where each has it's own color if Z is different.
            /// </summary>
            public void AddMultiColorLine(cPoint3D i_Start, cPoint3D i_End, int s32_Parts)
            {
                if (mi_ColorScheme == null)
                    throw new Exception("This function does not make sense if you do not specify a ColorScheme");

                // The line will have only one color because the color is calculated from the Z value only.
                if (i_Start.md_Z == i_End.md_Z)
                {
                    AddSolidLine(i_Start, i_End, null);
                    return;
                }

                double d_X = i_Start.md_X;
                double d_Y = i_Start.md_Y;
                double d_Z = i_Start.md_Z;

                double d_DeltaX = (i_End.md_X - d_X) / s32_Parts;
                double d_DeltaY = (i_End.md_Y - d_Y) / s32_Parts;
                double d_DeltaZ = (i_End.md_Z - d_Z) / s32_Parts;

                cPoint i_Prev = new cPoint(i_Start, 1, ""); // show tooltip for start point
                for (int i=0; i<s32_Parts; i++)
                {
                    d_X += d_DeltaX;
                    d_Y += d_DeltaY;
                    d_Z += d_DeltaZ;
                    cPoint i_Point = new cPoint(d_X, d_Y, d_Z,  0, null); // no tooltip for intermediate lines
                    mi_Lines.Add(new cLine(i_Prev, i_Point, null));
                    i_Prev = i_Point;
                }

                cPoint i_Last = new cPoint(i_End, 1, ""); // show tooltip for end point
                mi_Lines.Add(new cLine(i_Prev, i_Last, null));
                ms32_Lines ++;
            }

            public override void CalcMinMax()
            {
                foreach (cLine i_Line in mi_Lines)
                {
                    foreach (cPoint i_Point in i_Line.mi_Points)
                    {
                        mi_Inst.mi_MinMax.AddPoint(i_Point.mi_P3D);
                    }
                }
            }

            public override void PreProcess()
            {
                foreach (cLine i_Line in mi_Lines)
                {
                    // Load missing Pens from ColorScheme which the user has not specified.
                    if (mi_ColorScheme != null)
                    {
                        double d_FactorZ = mi_Inst.mi_MinMax.CalcFactorZ(i_Line.md_AvrgZ);
                        int    s32_Index = mi_ColorScheme.CalcIndex(d_FactorZ);

                        if (i_Line.mi_Pen == null) 
                            i_Line.mi_Pen = mi_ColorScheme.GetPen(s32_Index);

                        i_Line.mi_Pen.StartCap = LineCap.Round;
                        i_Line.mi_Pen.EndCap   = LineCap.Round;
                    }

                    mi_Inst.mi_Tooltip.AddPoint(i_Line.mi_Points[0]); // Start
                    mi_Inst.mi_Tooltip.AddPoint(i_Line.mi_Points[1]); // End
                }

                mi_Inst.mi_MinMax.ms32_TotLines     += ms32_Lines;
                mi_Inst.mi_MinMax.ms32_TotLineParts += mi_Lines.Count;
            }

            public override void ProjectDrawObjects()
            {
                // Convert 3D --> 2D
                foreach (cLine i_Line in mi_Lines)
                {
                    i_Line.mi_Points[0].Project3D(mi_Inst);
                    i_Line.mi_Points[1].Project3D(mi_Inst);

                    i_Line.CalcSortOrder(mi_Inst);

                    mi_Inst.InsertDrawObject(i_Line);
                }
            }
        }

        #endregion

        #region cPolygonData

        public class cPolygonData : cRenderData
        {
            private cColorScheme   mi_ColorScheme;
            private eSurfaceMode   me_Mode;
            private List<cPolygon> mi_Polygons  = new List<cPolygon>();
            private Pen            mi_BorderPen = new Pen(Color.Black, 1);  

            // This Pen draws the thin black lines around the border of the filled polygons.
            // You can set the Pen to null and no black lines will be drawn.
            public Pen BorderPen
            {
                set { mi_BorderPen = value; }
            }

            public cPolygonData(eSurfaceMode e_Mode, cColorScheme i_ColorScheme)
            {
                mi_ColorScheme = i_ColorScheme;
                me_Mode        = e_Mode;
            }

            /// <summary>
            /// In contrast to other drawing libraries (like WPF or Direct3D) you can add polygons of any size here.
            /// A polygon can have from at least 3 up to any amount of edges.
            /// </summary>
            public void AddPolygon(List<cPoint3D> i_Points3D)
            {
                if (i_Points3D.Count < 3)
                    throw new ArgumentException();

                mi_Polygons.Add(new cPolygon(me_Mode, i_Points3D.ToArray()));
            }

            public override void CalcMinMax()
            {
                foreach (cPolygon i_Polygon in mi_Polygons)
                {
                    foreach (cPoint i_Point in i_Polygon.mi_Points)
                    {
                        mi_Inst.mi_MinMax.AddPoint(i_Point.mi_P3D);
                    }
                }
            }

            public override void PreProcess()
            {
                foreach (cPolygon i_Polygon in mi_Polygons)
                {
                    // In mode eSurfaceMode.Fill tooltips may be confusing.
                    // When rendering a 3D object all the points in the back will also create a tooltip although they are invisible.
                    // Please read the documentation!
                    foreach (cPoint i_Point in i_Polygon.mi_Points)
                    {
                        mi_Inst.mi_Tooltip.AddPoint(i_Point);
                    }

                    cPolygon2D i_Poly2D = i_Polygon.mi_Polygon2D;

                    double  d_FactorZ = mi_Inst.mi_MinMax.CalcFactorZ(i_Poly2D.md_AvrgZ);
                    int     s32_Index = mi_ColorScheme.CalcIndex(d_FactorZ);
                    i_Poly2D.mi_Brush = mi_ColorScheme.GetBrush (s32_Index);

                    if (me_Mode == eSurfaceMode.Fill)
                        i_Poly2D.mi_Pen = mi_BorderPen; // thin black
                    else
                        i_Poly2D.mi_Pen = mi_ColorScheme.GetPen(s32_Index); // user defined
                }

                mi_Inst.mi_MinMax.ms32_TotPolygons += mi_Polygons.Count;
            }

            public override void ProjectDrawObjects()
            {
                // Convert 3D --> 2D
                foreach (cPolygon i_Polygon in mi_Polygons)
                {
                    i_Polygon.Project3D(mi_Inst);

                    i_Polygon.mi_Polygon2D.CalcSortOrder(mi_Inst);

                    mi_Inst.InsertDrawObject(i_Polygon.mi_Polygon2D);
                }
            }
        }

        #endregion

        #region cMessgData

        public class cMessgData
        {
            private String  ms_Text;
            private Brush   mi_Brush;
            private int     ms32_PosX;
            private int     ms32_PosY;
            private Font    mi_Font;
            private SizeF   mk_Size;

            /// <summary>
            /// If X is negative, it is displayed right  aligned at X pixels from the right
            /// If Y is negative, it is displayed bottom aligned at Y pixels from the bottom
            /// </summary>
            public cMessgData(String s_Text, int X, int Y, Color c_Color,
                              FontStyle e_FontStyle = FontStyle.Bold, 
                              int     s32_FontSize  = 9, 
                              String    s_FontFace  = "Tahoma")
            {
                ms_Text   = s_Text;
                ms32_PosX = X;
                ms32_PosY = Y;
                mi_Brush  = new SolidBrush(c_Color);
                mi_Font   = new Font(s_FontFace, s32_FontSize, e_FontStyle);
            }

            /// <summary>
            /// Here you can change the text without loading all the render objects again.
            /// Afterwards call graph3D.Invalidate()
            /// </summary>
            public void Update(String s_NewText, Color c_NewColor)
            {
                ms_Text  = s_NewText;
                mi_Brush = new SolidBrush(c_NewColor);
                mk_Size  = SizeF.Empty;
            }

            public void Draw(Graphics i_Graph, Rectangle k_Client)
            {
                float X = ms32_PosX;
                float Y = ms32_PosY;

                if (X < 0 || Y < 0)
                {
                    // Speed optimization: Measure the size only once.
                    if (mk_Size.IsEmpty)
                        mk_Size = i_Graph.MeasureString(ms_Text, mi_Font);

                    if (X < 0) X += k_Client.Width  - mk_Size.Width;
                    if (Y < 0) Y += k_Client.Height - mk_Size.Height;
                }

                i_Graph.DrawString(ms_Text, mi_Font, mi_Brush, X, Y);
            }
        }

        #endregion

        // ================= PRIVATE =================

        #region cPoint2D

        private class cPoint2D
        {
            public double md_X;
            public double md_Y;

            public cPoint2D()
            {
            }

            public cPoint2D(double X, double Y)
            {
                md_X = X;
                md_Y = Y;
            }

            public PointF Coord
            {
                get { return new PointF((float)md_X, (float)md_Y); }
            }

            public bool IsValid
            {
                get
                {
                    // The screen will always be smaller than 9999 pixels
                    return (!Double.IsNaN(md_X) && Math.Abs(md_X) < 9999.9 &&
                            !Double.IsNaN(md_Y) && Math.Abs(md_Y) < 9999.9);
                }
            }

            // For debugging in Visual Studio
            public override string ToString()
            {
                return String.Format("{0:0.00}, {1:0.00}", md_X, md_Y);
            }
        }

        #endregion

        #region cPolygon

        private class cPolygon
        {
            public cPoint[]   mi_Points; // 3D and 2D
            public cPolygon2D mi_Polygon2D;

            /// <summary>
            /// Constructor 1
            /// </summary>
            public cPolygon(eSurfaceMode e_Mode, params cPoint[] i_Points)
            {
                mi_Points    = i_Points;      
                mi_Polygon2D = new cPolygon2D(e_Mode, mi_Points);
            }

            /// <summary>
            /// Constructor 2
            /// </summary>
            public cPolygon(eSurfaceMode e_Mode, params cPoint3D[] i_Points3D)
            {
                mi_Points = new cPoint[i_Points3D.Length];
                for (int i=0; i<i_Points3D.Length; i++)
                {
                    mi_Points[i] = new cPoint(i_Points3D[i], 1, "");
                }
                mi_Polygon2D = new cPolygon2D(e_Mode, mi_Points);
            }

            public void Project3D(Graph3D i_Inst)
            {
                foreach (cPoint i_Point in mi_Points)
                {
                    i_Point.Project3D(i_Inst);
                }
                mi_Polygon2D.LoadPoints(mi_Points);
            }

            /// <summary>
            /// For debugging in Visual Studio
            /// </summary>
            public override string ToString()
            {
                return String.Format("Polygon ({0} points)", mi_Points.Length);
            }
        }

        #endregion

        #region cPoint

        private class cPoint
        {
            public cPoint3D mi_P3D;
            public cPoint2D mi_P2D;
            public int      ms32_RadiusTip; // if 0 --> no tooltip
            public String   ms_UserTip;

            public cPoint()
            {
                mi_P3D = new cPoint3D();
                mi_P2D = new cPoint2D();
            }

            public cPoint(double X, double Y, double Z, int s32_RadiusTip, String s_UserTip)
              : this(new cPoint3D(X, Y, Z), s32_RadiusTip, s_UserTip)
            {
            }

            /// <summary>
            /// s32_RadiusTip > 0 --> a tooltip with the 3D coordinates X, Y, Z will be displayed.
            /// The radius defines at which distance of the mouse the tooltip pops up.
            /// s_UserTip is an optional additional user defined tooltip.
            /// </summary>
            public cPoint(cPoint3D i_Point3D, int s32_RadiusTip, String s_UserTip)
            {
                mi_P3D = i_Point3D;
                mi_P2D = new cPoint2D();

                if (s32_RadiusTip > 0)
                    ms32_RadiusTip = s32_RadiusTip + 6;
                
                if (s_UserTip != null)
                    ms_UserTip = s_UserTip.Trim();
            }

            public void Project3D(Graph3D i_Inst)
            {
                mi_P2D = i_Inst.mi_Transform.Project3D(mi_P3D, i_Inst.mi_MinMax.mi_Center3D);
            }

            /// <summary>
            /// For Tooltip
            /// </summary>
            public override string ToString()
            {
                String s_TT = String.Format("X = {0}\nY = {1}\nZ = {2}", FormatDouble(mi_P3D.md_X), 
                                                                         FormatDouble(mi_P3D.md_Y), 
                                                                         FormatDouble(mi_P3D.md_Z));
                if (ms_UserTip != null)
                    s_TT += "\n" + ms_UserTip;
                    
                return s_TT;
            }
        }

        #endregion

        #region cTooltip

        private class cTooltip
        {
            private Graph3D      mi_Inst    = null;
            private bool         mb_Enabled = true;
            private ToolTip      mi_Tooltip = new ToolTip();
            private List<cPoint> mi_Points  = new List<cPoint>();
            private cPoint       mi_Last    = null;

            public bool Enabled
            {
                get { return mb_Enabled;  }
                set { mb_Enabled = value; }
            }

            // Constructor
            public cTooltip(Graph3D i_Inst)
            {
                mi_Inst = i_Inst;

                mi_Tooltip.AutoPopDelay = 30000;
                mi_Tooltip.InitialDelay = 50;
                mi_Tooltip.ReshowDelay  = 50;
            }

            public void Clear()
            {
                mi_Points.Clear();
                mi_Last = null;
                Hide();
            }

            /// <summary>
            /// ATTENTION: The tooltip will only work if you call later i_Point.Project() 
            /// which adds the 2 dimensional coordinates in i_Point.mi_P2D !
            /// </summary>
            public void AddPoint(cPoint i_Point)
            {
                if (i_Point != null && i_Point.ms32_RadiusTip > 0)
                    mi_Points.Add(i_Point);
            }

            public void Hide()
            {
                mi_Tooltip.Hide(mi_Inst);
            }

            public void OnMouseMove(MouseEventArgs e)
            {
                if (!mb_Enabled)
                {
                    Hide();
                    return;
                }

                int s32_MouseX  = e.X - mi_Inst.mi_Mouse.mk_OffMove.X - mi_Inst.mi_Mouse.mk_OffCoord.X;
                int s32_MouseY  = e.Y - mi_Inst.mi_Mouse.mk_OffMove.Y - mi_Inst.mi_Mouse.mk_OffCoord.Y;
                int s32_MinDist = int.MaxValue;

                cPoint i_Nearest = null;
                foreach (cPoint i_Point in mi_Points)
                {
                    int s32_DiffX = (int)i_Point.mi_P2D.md_X - s32_MouseX;
                    int s32_DiffY = (int)i_Point.mi_P2D.md_Y - s32_MouseY;
                    int s32_Dist  = (int)Math.Sqrt(s32_DiffX * s32_DiffX + s32_DiffY * s32_DiffY);
                    if (s32_Dist < s32_MinDist)
                    {
                        s32_MinDist = s32_Dist;
                        i_Nearest   = i_Point;
                    }
                }

                if (i_Nearest != null && s32_MinDist < i_Nearest.ms32_RadiusTip)
                {
                    if (mi_Last != i_Nearest)
                    {
                        mi_Last = i_Nearest;
                        mi_Tooltip.Show(i_Nearest.ToString(), mi_Inst, e.X + 10, e.Y + 10);
                    }
                }
                else
                {
                    mi_Last = null;
                    Hide();
                }
            }
        }

        #endregion

        #region cAxis

        private class cAxis
        {
            public Pen         mi_RasterPen;   // Raster lines (brigther)
            public Pen         mi_AxisPen;     // Main coordinate axis (darker)
            public SolidBrush  mi_LegendBrush; // Label text
            public String      ms_LegendText; 

            public cAxis(Color c_Color)
            {
                SetColor(c_Color);
            }
            public void SetColor(Color c_Color)
            {
                mi_LegendBrush = new SolidBrush(c_Color); 
                mi_AxisPen     = new Pen(c_Color, 3);     
                mi_RasterPen   = new Pen(BrightenColor(c_Color), 1);
            }
        }

        #endregion

        // ----- DrawObjects -----

        #region cDrawObj

        /// <summary>
        /// Base class for cLine, cScatter, cPolygon
        /// </summary>
        private class cDrawObj
        {
            public double        md_Sort;  // sorting is important. Always draw from back to front.
            public double        md_AvrgX; // 3D center point of draw object
            public double        md_AvrgY; // 3D center point of draw object
            public double        md_AvrgZ; // 3D center point of draw object
            public SmoothingMode me_SmoothMode;

            public virtual bool IsValid
            {
                get { throw new NotImplementedException(); }
            }

            public virtual void Render(Graphics i_Graph)
            {
                throw new NotImplementedException();
            }

            /// <summary>
            /// Uses the center of a draw object (e.g. the middle of a line) to calculate in which order the draw objects are rendered.
            /// This is called for all user defined objects, but not for the coordinate system which has it's own logic.
            /// </summary>
            public void CalcSortOrder(Graph3D i_Inst)
            {
                // cPolygon2D, cLine and cScatter must be rendered in correct order: from back to front, otherwise drawing artifacts.
                // What is back and what is front depends on Theta and Phi.
                md_Sort = i_Inst.mi_Transform.ProjectXY(md_AvrgX - i_Inst.mi_MinMax.md_MinX,
                                                        md_AvrgY - i_Inst.mi_MinMax.md_MinY,
                                                        md_AvrgZ - i_Inst.mi_MinMax.md_MinZ);
            }
        }

        #endregion

        #region cLine

        private class cLine : cDrawObj
        {
            public eCoord   me_Line   = eCoord.Invalid;  // main coordinate
            public eCoord   me_Offset = eCoord.Invalid;  // secondary coordinate
            public String   ms_Label;                    // Label for axis
            public Pen      mi_Pen;
            public double   md_Angle;
            public cPoint[] mi_Points = new cPoint[2] { new cPoint(), new cPoint() }; // start and end point of line

            /// <summary>
            /// Only used for axis, does not need Avrg values
            /// </summary>
            public cLine()
            {
                me_SmoothMode = SmoothingMode.AntiAlias;
            }

            public cLine(cPoint i_Start, cPoint i_End, Pen i_Pen)
            {
                me_SmoothMode = SmoothingMode.AntiAlias;
                mi_Points[0]  = i_Start;
                mi_Points[1]  = i_End;
                mi_Pen        = i_Pen;

                md_AvrgX = (i_Start.mi_P3D.md_X + i_End.mi_P3D.md_X) / 2.0;
                md_AvrgY = (i_Start.mi_P3D.md_Y + i_End.mi_P3D.md_Y) / 2.0;
                md_AvrgZ = (i_Start.mi_P3D.md_Z + i_End.mi_P3D.md_Z) / 2.0;
            }

            public bool CoordEquals(cLine i_Line)
            {
                return mi_Points[0].mi_P3D.Equals(i_Line.mi_Points[0].mi_P3D) &&
                       mi_Points[1].mi_P3D.Equals(i_Line.mi_Points[1].mi_P3D);
            }

            // Graphics.DrawLine() with huge length may take several seconds and block the GUI thread --> hanging.
            public override bool IsValid
            {
                get { return mi_Points[0].mi_P2D.IsValid && mi_Points[1].mi_P2D.IsValid; }
            }

            /// <summary>
            /// Calculate the angle of the 3 main axis on the screen in a range from 0 to 360.
            /// </summary>
            public void CalcAngle2D()
            {
                double d_DX = mi_Points[1].mi_P2D.md_X - mi_Points[0].mi_P2D.md_X;
                double d_DY = mi_Points[1].mi_P2D.md_Y - mi_Points[0].mi_P2D.md_Y;
                md_Angle = Math.Atan2(d_DY, d_DX) * 180.0 / Math.PI;
                if (md_Angle < 0.0) md_Angle += 360.0;
            }

            // For debugging in Visual Studio
            public override string ToString()
            {
                return String.Format("Line {{ {0} }} to {{ {1} }} ({2}, {3})", mi_Points[0].mi_P3D, mi_Points[1].mi_P3D, me_Line, me_Offset);
            }
        }

        #endregion

        #region cPolygon2D

        private class cPolygon2D : cDrawObj
        {
            public eSurfaceMode me_Mode;
            public PointF[]     mk_Screen;  // the polygon points in screen coordinates
            public Brush        mi_Brush;
            public Pen          mi_Pen;     // the pen to draw the lines 

            // Do not draw ploygons which are invalid
            private bool mb_Valid;
            public override bool IsValid
            {
                get { return mb_Valid; }
            }

            /// <summary>
            /// Constructor
            /// </summary>
            public cPolygon2D(eSurfaceMode e_Mode, params cPoint[] i_Poly)
            {
                me_Mode   = e_Mode;
                mk_Screen = new PointF[i_Poly.Length];

                // Drawing polygon lines with antialias makes them very thick and black.
                // Do not smooth the lines when the polygons are filled with color and the lines are only separators.
                // But use smooth mode if only the lines are drawn.
                me_SmoothMode = (e_Mode == eSurfaceMode.Lines) ? SmoothingMode.AntiAlias : SmoothingMode.None;

                foreach (cPoint i_Point in i_Poly)
                {
                    md_AvrgX += i_Point.mi_P3D.md_X;
                    md_AvrgY += i_Point.mi_P3D.md_Y;
                    md_AvrgZ += i_Point.mi_P3D.md_Z;
                }

                md_AvrgX /= i_Poly.Length;
                md_AvrgY /= i_Poly.Length;
                md_AvrgZ /= i_Poly.Length;
            }

            public void LoadPoints(cPoint[] i_Points)
            {
                Debug.Assert(i_Points.Length == mk_Screen.Length);
                mb_Valid = true;

                for (int i=0; i<i_Points.Length; i++)
                {
                    if (i_Points[i].mi_P2D.IsValid)
                        mk_Screen[i] = i_Points[i].mi_P2D.Coord;                       
                    else
                        mb_Valid = false;
                }
            }

            /// <summary>
            /// i_LinePen is normally a black Pen with 1 pixel width drawn between the polygons.
            /// </summary>
            public override void Render(Graphics i_Graph)
            {
                if (me_Mode == eSurfaceMode.Fill)
                {
                    i_Graph.FillPolygon(mi_Brush, mk_Screen);
                }

                if (mi_Pen != null)
                {
                    // ATTENTION: DrawPolygon() with a Pen > 1 pixel is buggy (artifacts)!
                    int T = mk_Screen.Length - 1;
                    for (int F=0; F<mk_Screen.Length; F++)
                    {
                        i_Graph.DrawLine(mi_Pen, mk_Screen[F], mk_Screen[T]);
                        T = F;
                    }    
                }
            }

            /// <summary>
            /// For debugging in Visual Studio
            /// </summary>
            public override string ToString()
            {
                return String.Format("Polygon2D ({0} points)", mk_Screen.Length);
            }
        }

        #endregion

        #region cScatter

        private class cScatter : cDrawObj
        {
            public cPoint        mi_Point;
            public PointF        mk_TopLeft; // top left corner in screen coordinates
            public Brush         mi_Brush;
            public Pen           mi_Pen;
            public cScatter      mi_Previous;
            public int           ms32_Diameter;
            public eScatterShape me_Shape;
            public Point[]       mk_Polygon; // triangle

            // Do not draw scatter shapes or lines which are invalid
            private bool mb_Valid;
            public override bool IsValid
            {
                get { return mb_Valid; }
            }

            /// <summary>
            /// Constructor for Shapes
            /// i_Brush == null --> use Brush from ColorScheme
            /// i_Brush != null --> use i_Brush
            /// s_UserTip is an optional additional user defined tooltip
            /// </summary>
            public cScatter(double X, double Y, double Z, eScatterShape e_Shape, int s32_Radius, Brush i_Brush, String s_UserTip = "")
            {
                mi_Point      = new cPoint(X, Y, Z, s32_Radius, s_UserTip);
                me_Shape      = e_Shape;
                ms32_Diameter = Math.Max(1, s32_Radius) * 2;
                mi_Brush      = i_Brush;
                me_SmoothMode = SmoothingMode.AntiAlias;

                md_AvrgX = X;
                md_AvrgY = Y;
                md_AvrgZ = Z;
            }

            /// <summary>
            /// Constructor for Lines
            /// i_Pen == null --> use Pen from ColorScheme
            /// i_Pen != null --> use i_Pen
            /// s_UserTip is an optional additional user defined tooltip
            /// </summary>
            public cScatter(double X, double Y, double Z, Pen i_Pen, int s32_Radius, String s_UserTip = "")
            {
                mi_Point      = new cPoint(X, Y, Z, s32_Radius, s_UserTip);
                me_Shape      = (eScatterShape) 0; // Invalid
                mi_Pen        = i_Pen;
                me_SmoothMode = SmoothingMode.AntiAlias;

                md_AvrgX = X;
                md_AvrgY = Y;
                md_AvrgZ = Z;
            }

            public void SetPoint2D(cPoint2D i_Point2D)
            {
                mk_TopLeft = i_Point2D.Coord;
                mb_Valid   = i_Point2D.IsValid;

                // Move coordinate from center to upper left corner of circle
                mk_TopLeft.X -= ms32_Diameter / 2;
                mk_TopLeft.Y -= ms32_Diameter / 2;

                switch (me_Shape)
                {
                    case eScatterShape.Triangle:
                        mk_Polygon = new Point[3];
                        // top center
                        mk_Polygon[0].X = (int)mk_TopLeft.X + ms32_Diameter / 2;
                        mk_Polygon[0].Y = (int)mk_TopLeft.Y;
                        // bottom left
                        mk_Polygon[1].X = (int)mk_TopLeft.X;
                        mk_Polygon[1].Y = (int)mk_TopLeft.Y + ms32_Diameter;
                        // bottom right
                        mk_Polygon[2].X = (int)mk_TopLeft.X + ms32_Diameter;
                        mk_Polygon[2].Y = (int)mk_TopLeft.Y + ms32_Diameter;
                        break;
                }
            }

            public override void Render(Graphics i_Graph)
            {
                switch (me_Shape)
                {
                    case eScatterShape.Circle:
                        i_Graph.FillEllipse(mi_Brush, mk_TopLeft.X, mk_TopLeft.Y, ms32_Diameter, ms32_Diameter);
                        break;

                    case eScatterShape.Square:
                        i_Graph.FillRectangle(mi_Brush, mk_TopLeft.X, mk_TopLeft.Y, ms32_Diameter, ms32_Diameter);
                        break;

                    case eScatterShape.Triangle:
                        i_Graph.FillPolygon(mi_Brush, mk_Polygon);
                        break;

                    default: // me_Shape is invalid --> draw line
                        if (mi_Previous != null)
                            i_Graph.DrawLine(mi_Pen, mi_Previous.mk_TopLeft, mk_TopLeft);
                        break;
                }
            }

            /// <summary>
            /// For Debugging in Visual Studio
            /// </summary>
            public override string ToString()
            {
                return String.Format("Scatter {0} at {1}, Diameter {2}", me_Shape, mi_Point, ms32_Diameter);
            }
        }

        #endregion

        // ----- Math Stuff ------

        #region cMouse 

        private class cMouse
        {
            public eMouseAction me_Action;     // left mouse button action
            public Point        mk_LastPos;    // last mouse location
            public Point        mk_OffMove;    // Mouse offset after moving the graph with the mouse
            public Point        mk_OffCoord;   // Offset caused by labels in coordinate system
            public TrackBar     mi_TrackRho;   // Rho trackbar (optional)
            public TrackBar     mi_TrackTheta; // Theta trackbar (optional)
            public TrackBar     mi_TrackPhi;   // Phi trackbar (optional)
            public double       md_Rho   = VALUES_RHO  [2]; // 2 = Default value
            public double       md_Theta = VALUES_THETA[2]; // 2 = Default value
            public double       md_Phi   = VALUES_PHI  [2]; // 2 = Default value

            public void AssignTrackbar(eMouseAction e_Trackbar, TrackBar i_Trackbar, EventHandler i_OnScroll)
            {
                if (i_Trackbar == null)
                    return;

                double[] d_Values = null;
                switch (e_Trackbar)
                {
                    case eMouseAction.Rho:
                        d_Values      = VALUES_RHO;
                        mi_TrackRho   = i_Trackbar;
                        break;
                    case eMouseAction.Theta:
                        d_Values      = VALUES_THETA;
                        mi_TrackTheta = i_Trackbar;
                        break;
                    case eMouseAction.Phi:
                        d_Values      = VALUES_PHI;
                        mi_TrackPhi   = i_Trackbar;
                        break;
                }

                i_Trackbar.Minimum = (int)d_Values[0]; // 0 = Minimum
                i_Trackbar.Maximum = (int)d_Values[1]; // 1 = Maximum
                i_Trackbar.Value   = (int)d_Values[2]; // 2 = Default value
                i_Trackbar.Scroll += i_OnScroll;
            }

            /// <summary>
            /// User has moved the TrackBar
            /// </summary>
            public void OnTrackBarScroll()
            {
                if (mi_TrackRho   != null) md_Rho   = mi_TrackRho  .Value;
                if (mi_TrackTheta != null) md_Theta = mi_TrackTheta.Value;
                if (mi_TrackPhi   != null) md_Phi   = mi_TrackPhi  .Value;
            }

            public bool OnMouseWheel(int s32_Delta)
            {
                if (me_Action != eMouseAction.None)
                    return false;

                me_Action = eMouseAction.Rho;
                OnMouseMove(0, s32_Delta / 10);
                me_Action = eMouseAction.None;
                return true;
            }

            /// <summary>
            /// User has dragged the mouse over the 3D control
            /// </summary>
            public void OnMouseMove(int s32_DiffX, int s32_DiffY)
            {
                switch (me_Action)
                {
                    case eMouseAction.Rho:
                        md_Rho += s32_DiffY * VALUES_RHO[3];  // 3 = Factor
                        SetRho(md_Rho);
                        break;

                    case eMouseAction.Theta:
                        md_Theta -= s32_DiffY * VALUES_THETA[3];  // 3 = Factor
                        SetTheta(md_Theta);
                        break;

                    case eMouseAction.Phi:
                        md_Phi -= s32_DiffX * VALUES_PHI[3]; // 3 = Factor
                        SetPhi(md_Phi);
                        break;
                }
            }

            public void SetRho(double d_Rho)
            {
                md_Rho = d_Rho;
                md_Rho = Math.Max(md_Rho, VALUES_RHO[0]); // 0 = Minimum
                md_Rho = Math.Min(md_Rho, VALUES_RHO[1]); // 1 = Maximum
                if (mi_TrackRho != null)
                    mi_TrackRho.Value = (int)md_Rho;
            }
            public void SetTheta(double d_Theta)
            {
                md_Theta = d_Theta;
                md_Theta = Math.Max(md_Theta, VALUES_THETA[0]); // 0 = Minimum
                md_Theta = Math.Min(md_Theta, VALUES_THETA[1]); // 1 = Maximum
                if (mi_TrackTheta != null)
                    mi_TrackTheta.Value = (int)md_Theta;
            }
            public void SetPhi(double d_Phi)
            {
                md_Phi = d_Phi;
                if (md_Phi > 360.0) md_Phi -= 360.0; // continuous rotation
                if (md_Phi <   0.0) md_Phi += 360.0; // continuous rotation
                if (mi_TrackPhi != null)
                    mi_TrackPhi.Value = (int)md_Phi;
            }
        }

        #endregion

        #region cMinMax3D

        private class cMinMax3D
        {
            // Center point is subtracted from 3D values before converting to 2D
            public  cPoint3D mi_Center3D;
            public  double   md_MinX, md_MaxX, md_MinY, md_MaxY, md_MinZ, md_MaxZ;
            public  int      ms32_TotPoints;
            public  int      ms32_TotLines;
            public  int      ms32_TotLineParts;
            public  int      ms32_TotPolygons;
            private bool     mb_Empty = true;

            public bool IsEmpty
            {
                get { return mb_Empty; }
            }

            /// <summary>
            /// To be displayed to the user
            /// </summary>
            public String ObjectCount
            {
                get
                {
                    StringBuilder i_Out = new StringBuilder();
                    if (ms32_TotPoints   > 0) i_Out.AppendLine(ms32_TotPoints   + " points");
                    if (ms32_TotPolygons > 0) i_Out.AppendLine(ms32_TotPolygons + " polygons");
                    if (ms32_TotLines > 0)
                    {
                        if (ms32_TotLineParts > ms32_TotLines)
                            i_Out.AppendLine(ms32_TotLines + " lines, " + ms32_TotLineParts + " parts");
                        else
                            i_Out.AppendLine(ms32_TotLines + " lines");
                    }
                    return i_Out.ToString().Trim();
                }
            }

            public void Reset()
            {
                md_MinX = double.PositiveInfinity;
                md_MaxX = double.NegativeInfinity;
                md_MinY = double.PositiveInfinity;
                md_MaxY = double.NegativeInfinity;
                md_MinZ = double.PositiveInfinity;
                md_MaxZ = double.NegativeInfinity;

                mb_Empty          = true;
                mi_Center3D       = new cPoint3D();
                ms32_TotPoints    = 0;
                ms32_TotLines     = 0;
                ms32_TotLineParts = 0;
                ms32_TotPolygons  = 0;
            }

            public void AddPoint(cPoint3D i_Point)
            {
                md_MinX = Math.Min(md_MinX, i_Point.md_X);
                md_MaxX = Math.Max(md_MaxX, i_Point.md_X);
                md_MinY = Math.Min(md_MinY, i_Point.md_Y);
                md_MaxY = Math.Max(md_MaxY, i_Point.md_Y);
                md_MinZ = Math.Min(md_MinZ, i_Point.md_Z);
                md_MaxZ = Math.Max(md_MaxZ, i_Point.md_Z);
                mb_Empty = false;
            }

            public double CalcFactorZ(double d_Value)
            {
                return (d_Value - md_MinZ) / (md_MaxZ - md_MinZ);
            }
        }

        #endregion

        #region cQuadrant

        private class cQuadrant
        {
            public double md_SortXY;   // Sort order of raster in area XY  (red)
            public double md_SortXZ;   // Sort order of X axis and raster in area XZ (blue)
            public double md_SortYZ;   // Sort order of Y axis and raster in area YZ (green)
            public int    ms32_Quadrant;
            public bool   mb_BottomView;

            public void Calculate(double d_Phi, cLine i_AxisX, cLine i_AxisY, cLine i_AxisZ)
            {
                // Split rotation into 4 sections (0...3) which increment every 90° starting at 45°
                int s32_Section45 = (int)d_Phi + 45;
                if (s32_Section45 > 360) s32_Section45 -= 360;
                s32_Section45 = Math.Min(3, s32_Section45 / 90);

                // Theta elevation lets the camera watch the graph from the top or bottom
                switch (s32_Section45)
                {
                    case 0: mb_BottomView = i_AxisX.md_Angle < 180.0; break;
                    case 1: mb_BottomView = i_AxisY.md_Angle < 180.0; break;
                    case 2: mb_BottomView = i_AxisX.md_Angle > 180.0; break;
                    case 3: mb_BottomView = i_AxisY.md_Angle > 180.0; break;
                }

                // The quadrant changes when the 2D transformed Z axis is in line with the X or Y axis
                if (mb_BottomView)
                {
                    switch (s32_Section45)
                    {
                        case 0: ms32_Quadrant = i_AxisX.md_Angle + 180.0 < i_AxisZ.md_Angle ? 1 : 0; break;
                        case 1: ms32_Quadrant = i_AxisY.md_Angle + 180.0 < i_AxisZ.md_Angle ? 2 : 1; break;
                        case 2: ms32_Quadrant = i_AxisX.md_Angle         < i_AxisZ.md_Angle ? 3 : 2; break;
                        case 3: ms32_Quadrant = i_AxisY.md_Angle         < i_AxisZ.md_Angle ? 0 : 3; break;
                    }
                }
                else // Top View
                {
                    switch (s32_Section45)
                    {
                        case 0: ms32_Quadrant = i_AxisX.md_Angle         > i_AxisZ.md_Angle ? 1 : 0; break;
                        case 1: ms32_Quadrant = i_AxisY.md_Angle         > i_AxisZ.md_Angle ? 2 : 1; break;
                        case 2: ms32_Quadrant = i_AxisX.md_Angle + 180.0 > i_AxisZ.md_Angle ? 3 : 2; break;
                        case 3: ms32_Quadrant = i_AxisY.md_Angle + 180.0 > i_AxisZ.md_Angle ? 0 : 3; break;
                    }
                }

                md_SortXY = (mb_BottomView) ? 99999.9 : -99999.9;
                md_SortXZ = (ms32_Quadrant == 1 || ms32_Quadrant == 2) ? 99999.9 : -99999.9;
                md_SortYZ = (ms32_Quadrant == 0 || ms32_Quadrant == 1) ? 99999.9 : -99999.9;

                i_AxisX.md_Sort = md_SortXZ;
                i_AxisY.md_Sort = md_SortYZ;
                i_AxisZ.md_Sort = (ms32_Quadrant == 3) ? -99999.9 : 99999.9;

                // Debug.WriteLine(String.Format("Section: {0}  Quadrant: {1}", s32_Section45, ms32_Quadrant));
            }
        }

        #endregion

        #region cTransform

        private class cTransform
        {
            private double md_Dist; // Screen Distance
            private double md_sf;   // sf = sinus fi
            private double md_st;   // st = sinus theta
            private double md_cf;   // cf = cosinus fi
            private double md_ct;   // ct = cosinus theta
            private double md_Rho;
            // ----------------
            private double md_FactX;
            private double md_OffsX;
            private double md_FactY;
            private double md_OffsY;
            // ----------------
            public double  md_NormalizeX;
            public double  md_NormalizeY;
            public double  md_NormalizeZ;
            // ----------------
            Size mk_InitialSize = Size.Empty;

            public void SetCoefficients(cMouse i_Mouse)
            {
                md_Rho         =  i_Mouse.md_Rho;                           // Distance of viewer (zoom)
                double d_Theta =  i_Mouse.md_Theta       * Math.PI / 180.0; // Height   of viewer (elevation)
                double d_Phi   = (i_Mouse.md_Phi -180.0) * Math.PI / 180.0; // Rotation around center (-pi ... +pi)

                md_sf   = Math.Sin(d_Phi);  
                md_cf   = Math.Cos(d_Phi);  
                md_st   = Math.Sin(d_Theta); // Theta = 0...pi --> st = 0 .. 1 .. 0
                md_ct   = Math.Cos(d_Theta); // Theta = 0...pi --> ct = 1 .. 0 .. -1
                md_Dist = 0.5; // Camera distance. Smaller values result in ugly stretched egdes when rotating.
            }

            /// <summary>
            /// The initial size is needed to calculate the user resizing factor.
            /// To assure that it is correct it must be set when the control has already been created.
            /// Then it will be the size that was defined in Visual Studio Form Designer.
            /// </summary>
            public void SetInitialSize(Size k_Size)
            {
                mk_InitialSize = k_Size;
                SetSize(k_Size);
            }

            /// <summary>
            /// The control has been resized.
            /// This may be called with an invalid size before the control is created!
            /// </summary>
            public void SetSize(Size k_Size) // Control.ClientSize
            {
                if (mk_InitialSize == Size.Empty)
                    return;

                double d_Width  = k_Size.Width  * 0.0254 / 96.0; // 0.0254 meter = 1 inch. Screen has 96 DPI
                double d_Height = k_Size.Height * 0.0254 / 96.0;

                // linear transformation coefficients
                md_FactX =  k_Size.Width  / d_Width;
                md_FactY = -k_Size.Height / d_Height;
               
                md_OffsX =  md_FactX * d_Width  / 2.0;
                md_OffsY = -md_FactY * d_Height / 2.0;

                // -----------------------------------

                double d_ResizeX = (double)k_Size.Width  / mk_InitialSize.Width;
                double d_ResizeY = (double)k_Size.Height / mk_InitialSize.Height;
                double d_Resize  = Math.Min(d_ResizeX, d_ResizeY);

                md_FactX *= d_Resize;
                md_FactY *= d_Resize;
            }

            // Performs projection. Calculates screen coordinates for 3D point.
            // returns Point in 2D space of the screen.
            public cPoint2D Project3D(cPoint3D i_Point3D, cPoint3D i_Center3D)
            {
                double X = (i_Point3D.md_X - i_Center3D.md_X) * md_NormalizeX;
                double Y = (i_Point3D.md_Y - i_Center3D.md_Y) * md_NormalizeY;
                double Z = (i_Point3D.md_Z - i_Center3D.md_Z) * md_NormalizeZ;

                // 3D coordinates with center point in the middle of the screen
                // X positive to the right, X negative to the left
                // Y positive to the top,   Y negative to the bottom
                double xn = -md_sf *         X + md_cf         * Y;
                double yn = -md_cf * md_ct * X - md_sf * md_ct * Y + md_st * Z;
                double zn = -md_cf * md_st * X - md_sf * md_st * Y - md_ct * Z + md_Rho;

                zn = Math.Max(zn, 0.01);

                // Thales' theorem
                return new cPoint2D(md_FactX * xn * md_Dist / zn + md_OffsX,
                                    md_FactY * yn * md_Dist / zn + md_OffsY);
            }

            // Required for correct painting order of polygons (always from back to front)
            public double ProjectXY(double X, double Y, double Z = 0.0)
            {
                return X * md_cf + Y * md_sf + Z * md_ct;
            }
        }

        #endregion

        // Limits and default values for mouse actions and trackbars.
        // ATTENTION: It is strongly recommended not to change the MIN, MAX values.
        // The mouse factor defines how much mouse movement you need for a change.
        // A movement of mouse by approx 1000 pixels on the screen results in getting from Min to Max or vice versa.
        //
        //                                                     Min     Max   Default  MouseFactor
        static readonly double[] VALUES_RHO   = new double[] { 300,   1800,  1350,    2    };
        static readonly double[] VALUES_THETA = new double[] {  10,    170,    70,    0.25 }; // degree
        static readonly double[] VALUES_PHI   = new double[] {   0,    360,   230,    0.4  }; // degree  (continuous rotation)

        // The axis are 10% longer than the highest X,Y,Z value
        const double AXIS_EXCESS = 1.1;

        // For any strange reason the graph is not centered vertically
        const int VERT_OFFSET = -30;

        // Calculate 3-dimensional Z value from X,Y values
        public delegate double delRendererFunction(double X, double Y);

        Pen                mi_BorderPen      = null;
        SolidBrush         mi_TopLegendBrush = null;
        eRaster            me_Raster         = eRaster.Off;
        cAxis[]            mi_Axis           = new cAxis[3];
        cMouse             mi_Mouse          = new cMouse();
        cTooltip           mi_Tooltip;
        // --------------- User data -------------------------
        bool               mb_Updating       = false;        
        List<cMessgData>   mi_MessageData    = new List<cMessgData>();
        List<cRenderData>  mi_RenderData     = new List<cRenderData>();
        // --------------- Calculated data -------------------
        List<cDrawObj>     mi_DrawObjects    = new List<cDrawObj>();
        cTransform         mi_Transform      = new cTransform();
        cMinMax3D          mi_MinMax         = new cMinMax3D();
        cQuadrant          mi_Quadrant       = new cQuadrant();

        #region Properties 

        /// <summary>
        /// Shows a tooltip when the mouse is over a 3D point.
        /// This property can also be set in the Visual Studio Designer
        /// </summary>
        public bool TooltipEnabled
        {
            get { return mi_Tooltip.Enabled;  }
            set { mi_Tooltip.Enabled = value; }
        }

        /// <summary>
        /// Off:      draw no coordinate system
        /// MainAxis: draw solid axis X,Y,Z
        /// Raster:   draw axis and dotted raster lines
        /// Labels:   draw axis and dotted raster lines and labels in quadrant 3
        /// This property can also be set in the Visual Studio Designer
        /// </summary>
        public eRaster Raster
        {
            set
            {
                Debug.Assert(!InvokeRequired); // Do not call from other threads or use Invoke()
                if (me_Raster != value)
                {
                    me_Raster = value;
                    mi_DrawObjects.Clear(); // recalculate
                    Invalidate();           // repaint
                }
            }
            get
            {
                return me_Raster;
            }
        }

        /// <summary>
        /// setting BorderColor = Color.Gray draws a one pixel gray border around the control
        /// setting BorderColor = Color.Empty turns off the border
        /// This property can also be set in the Visual Studio Designer
        /// </summary>
        public Color BorderColor
        {
            set
            {
                Debug.Assert(!InvokeRequired); // Do not call from other threads or use Invoke()
                if (value.A == 0) mi_BorderPen = null; // transparent color
                else              mi_BorderPen = new Pen(value);
                Invalidate(); // repaint only
            }
            get
            {
                if (mi_BorderPen != null) return mi_BorderPen.Color;
                else                      return Color.Empty;
            }
        }

        /// <summary>
        /// Show a legend with Rotation, Elevation and Distance at the top left
        /// setting LegendColor = Color.Gray draws legend text in gray
        /// setting LegendColor = Color.Empty turns off the top legend
        /// This property can also be set in the Visual Studio Designer
        /// </summary>
        public Color TopLegendColor
        {
            set
            {
                Debug.Assert(!InvokeRequired); // Do not call from other threads or use Invoke()
                mi_TopLegendBrush = new SolidBrush(value);
                Invalidate(); // repaint only
            }
            get
            {
                if (mi_TopLegendBrush != null) return mi_TopLegendBrush.Color;
                else                           return Color.Empty;
            }
        }

        /// <summary>
        /// Set the colors of axis, raster lines and lables
        /// This property can also be set in the Visual Studio Designer
        /// </summary>
        public Color AxisX_Color
        {
            set { mi_Axis[(int)eCoord.X].SetColor(value); }
            get { return mi_Axis[(int)eCoord.X].mi_AxisPen.Color; }
        }
        public Color AxisY_Color
        {
            set { mi_Axis[(int)eCoord.Y].SetColor(value); }
            get { return mi_Axis[(int)eCoord.Y].mi_AxisPen.Color; }
        }
        public Color AxisZ_Color
        {
            set { mi_Axis[(int)eCoord.Z].SetColor(value); }
            get { return mi_Axis[(int)eCoord.Z].mi_AxisPen.Color; }
        }

        #endregion

        public void SetCoefficients(double d_Rho, double d_Theta, double d_Phi)
        {
            Debug.Assert(!InvokeRequired); // Do not call from other threads or use Invoke()

            mi_Mouse.SetRho  (d_Rho);
            mi_Mouse.SetTheta(d_Theta);
            mi_Mouse.SetPhi  (d_Phi);

            mi_Transform.SetCoefficients(mi_Mouse);

            mi_DrawObjects.Clear(); // recalculate
            Invalidate();           // repaint
        }

        /// <summary>
        /// returns the total count of loaded draw objects (points, lines or polygons)
        /// </summary>
        [ReadOnly(true)]
        [Browsable(false)]
        public String ObjectCount
        {
            get { return mi_MinMax.ObjectCount; }
        }

        /// <summary>
        /// Trackbars are optional for user interaction.
        /// If this function is never called thetrackbars are not used.
        /// </summary>
        public void AssignTrackBars(TrackBar i_Rho, TrackBar i_Theta, TrackBar i_Phi)
        {
            Debug.Assert(!InvokeRequired); // Do not call from other threads or use Invoke()
            mi_Mouse.AssignTrackbar(eMouseAction.Rho,   i_Rho,   new EventHandler(OnTrackbarScroll));
            mi_Mouse.AssignTrackbar(eMouseAction.Theta, i_Theta, new EventHandler(OnTrackbarScroll));
            mi_Mouse.AssignTrackbar(eMouseAction.Phi,   i_Phi,   new EventHandler(OnTrackbarScroll));
        }

        // ==================================================================================

        /// <summary>
        /// Constructor
        /// </summary>
        public Graph3D()
        {
            // avoid flicker
            SetStyle(ControlStyles.AllPaintingInWmPaint,  true);
            SetStyle(ControlStyles.OptimizedDoubleBuffer, true);

            // Load the default colors
            BackColor = Color.White;

            mi_Axis[(int)eCoord.X] = new cAxis(Color.DarkBlue);
            mi_Axis[(int)eCoord.Y] = new cAxis(Color.DarkGreen);
            mi_Axis[(int)eCoord.Z] = new cAxis(Color.DarkRed);

            mi_BorderPen      = new Pen       (Color.FromArgb(255,180,180,180), 1); // bright gray
            mi_TopLegendBrush = new SolidBrush(Color.FromArgb(255,200,200,150));    // beige

            mi_Transform.SetCoefficients(mi_Mouse);

            mi_Tooltip = new cTooltip(this);
        }

        // ------------------------------------------------------

        /// <summary>
        /// Correct Sequence: BeginUpdate(), AddRenderData(), EndUpdate()
        /// </summary>
        public void BeginUpdate()
        {
            mb_Updating = true;
            mi_MinMax.Reset();
            mi_MessageData.Clear();
            mi_RenderData .Clear();
            mi_DrawObjects.Clear();
            mi_Tooltip    .Clear();
            mi_Axis[(int)eCoord.X].ms_LegendText = null;
            mi_Axis[(int)eCoord.Y].ms_LegendText = null;
            mi_Axis[(int)eCoord.Z].ms_LegendText = null;
        }

        /// <summary>
        /// Correct Sequence: BeginUpdate(), AddRenderData(), EndUpdate()
        /// </summary>
        public void EndUpdate(eNormalize e_Normalize)
        {
            Debug.Assert(!InvokeRequired);
            if (!mb_Updating) throw new Exception("You must call BeginUpdate() first");

            foreach (cRenderData i_Data in mi_RenderData)
            {
                // Preprocess() needs mi_MinMax to have valid values for !ALL! 3D objects!
                i_Data.PreProcess();
            }

            NormalizeRanges(e_Normalize);

            mi_Mouse.mk_OffMove = Point.Empty;
            mb_Updating = false;

            Invalidate(); 
        }

        // ------------------------------------------------------

        /// <summary>
        /// Any legend can be null --> the label is not displayed
        /// </summary>
        public void SetAxisLegends(String s_LegendX, String s_LegendY, String s_LegendZ)
        {
            if (!mb_Updating) throw new Exception("You must call BeginUpdate() first");

            mi_Axis[(int)eCoord.X].ms_LegendText = s_LegendX;
            mi_Axis[(int)eCoord.Y].ms_LegendText = s_LegendY;
            mi_Axis[(int)eCoord.Z].ms_LegendText = s_LegendZ;
        }

        /// <summary>
        /// Adds a message to be shown, even if no 3D data is loaded
        /// </summary>
        public void AddMessageData(cMessgData i_Data)
        {
            if (!mb_Updating) throw new Exception("You must call BeginUpdate() first");

            mi_MessageData.Add(i_Data);
        }

        /// <summary>
        /// Here you can set a callback function which will be called to calculate the 3D values.
        /// Correct Sequence: BeginUpdate(), AddFunctionData(), EndUpdate()
        /// </summary>
        public void AddFunctionData(delRendererFunction f_Function, PointF k_Start, PointF k_End, double d_Density, 
                                    eSurfaceMode e_Surface, cColorScheme i_ColorScheme)
        {
            int s32_Cols = (int)((k_End.X - k_Start.X) / d_Density + 1);
            int s32_Rows = (int)((k_End.Y - k_Start.Y) / d_Density + 1);

            cSurfaceData i_Data = new cSurfaceData(e_Surface, s32_Cols, s32_Rows, i_ColorScheme);
            for (int C = 0; C < s32_Cols; C++)
            {
                double d_X = k_Start.X + d_Density * C;

                for (int R = 0; R < s32_Rows; R++)
                {
                    double d_Y = k_Start.Y + d_Density * R;
                    double d_Z = f_Function(d_X, d_Y);

                    i_Data.SetPointAt(C, R,  d_X, d_Y, d_Z);
                }
            }
            AddRenderData(i_Data);
        }

        /// <summary>
        /// Here you can add cSurfaceData or cScatterData
        /// Correct Sequence: BeginUpdate(), AddRenderData(), EndUpdate()
        /// </summary>
        public void AddRenderData(cRenderData i_Data)
        {
            if (!mb_Updating) throw new Exception("You must call BeginUpdate() first");

            i_Data.mi_Inst = this;
            mi_RenderData.Add(i_Data);
            i_Data.CalcMinMax();
        }

        // ========================================== PRIVATE ============================================

        /// <summary>
        /// This function normalizes the 3D ranges for the X,Y,Z coordinates.
        /// Otherwise a 3D range of X,Y from -10 to +10 will appear much smaller than a range from -100 to +100.
        /// It adapts the values so that rotation (phi) goes through the center of the X, Y pane.
        /// </summary>
        private void NormalizeRanges(eNormalize e_Normalize)
        {
            if (mi_MinMax.IsEmpty)
                return; // nothing loaded

            if (mi_MinMax.md_MaxX == mi_MinMax.md_MinX) { mi_MinMax.md_MinX -= 1.0; mi_MinMax.md_MaxX += 1.0; }
            if (mi_MinMax.md_MaxY == mi_MinMax.md_MinY) { mi_MinMax.md_MinY -= 1.0; mi_MinMax.md_MaxY += 1.0; }
            if (mi_MinMax.md_MaxZ == mi_MinMax.md_MinZ) { mi_MinMax.md_MinZ -= 1.0; mi_MinMax.md_MaxZ += 1.0; }

            // Normalize 3D X,Y values to compensate different ranges of X min ... X max and Y min ... Y max
            double d_RangeX = mi_MinMax.md_MaxX - mi_MinMax.md_MinX;
            double d_RangeY = mi_MinMax.md_MaxY - mi_MinMax.md_MinY;

            // Normalize 3D Z value to fit on screen
            double d_RangeZ;
            if (me_Raster == eRaster.Off)
                d_RangeZ = mi_MinMax.md_MaxZ - mi_MinMax.md_MinZ;
            else
                d_RangeZ = Math.Max(0, mi_MinMax.md_MaxZ) - Math.Min(0, mi_MinMax.md_MinZ);

            switch (e_Normalize)
            {
                case eNormalize.MaintainXY:
                    double d_RangeXY = (d_RangeX + d_RangeY) / 2; // average
                    d_RangeX = d_RangeXY;
                    d_RangeY = d_RangeXY;
                    break;

                case eNormalize.MaintainXYZ:
                    double d_RangeXYZ = (d_RangeX + d_RangeY + d_RangeZ) / 3; // average
                    d_RangeX = d_RangeXYZ;
                    d_RangeY = d_RangeXYZ;
                    d_RangeZ = d_RangeXYZ;
                    break;
            }

            mi_Transform.md_NormalizeX = 250.0 / d_RangeX;
            mi_Transform.md_NormalizeY = 250.0 / d_RangeY;
            mi_Transform.md_NormalizeZ = 250.0 / d_RangeZ; // RangeZ will never be zero.

            mi_MinMax.mi_Center3D.md_X = (mi_MinMax.md_MaxX + mi_MinMax.md_MinX) / 2.0;
            mi_MinMax.mi_Center3D.md_Y = (mi_MinMax.md_MaxY + mi_MinMax.md_MinY) / 2.0;

            if (me_Raster == eRaster.Off)
                mi_MinMax.mi_Center3D.md_Z = (mi_MinMax.md_MaxZ + mi_MinMax.md_MinZ) / 2.0;
            else
                mi_MinMax.mi_Center3D.md_Z = (Math.Max(0, mi_MinMax.md_MaxZ) + Math.Min(0, mi_MinMax.md_MinZ)) / 2.0;
        }

        private void CreateCoordinateSystem(Graphics i_Graph)
        {
            mi_Mouse.mk_OffCoord = new Point(0, VERT_OFFSET);

            if (me_Raster == eRaster.Off || mi_MinMax.IsEmpty)
                return;

            List<cLine> i_Lines = new List<cLine>();

            // Add the 3 coordinate system main axis
            for (int A=0; A<3; A++)
            {
                cLine i_Axis  = new cLine();
                i_Axis.mi_Pen = mi_Axis[A].mi_AxisPen;

                switch ((eCoord)A)
                {
                    case eCoord.X: // Blue
                        i_Axis.mi_Points[0].mi_P3D.md_X = Math.Min(0.0, mi_MinMax.md_MinX * AXIS_EXCESS);
                        i_Axis.mi_Points[1].mi_P3D.md_X = Math.Max(0.0, mi_MinMax.md_MaxX * AXIS_EXCESS);
                        // ---------------------------
                        i_Axis.mi_Points[0].mi_P3D.md_Y = Math.Min(0.0, mi_MinMax.md_MinY * AXIS_EXCESS); // X axis at minimum Y position
                        i_Axis.mi_Points[1].mi_P3D.md_Y = Math.Min(0.0, mi_MinMax.md_MinY * AXIS_EXCESS); // X axis at minimum Y position
                        i_Axis.me_Line   = eCoord.X;
                        i_Axis.me_Offset = eCoord.X; // Hide zero label at X axis (X,X invalid)
                        break;
                    case eCoord.Y: // Green
                        i_Axis.mi_Points[0].mi_P3D.md_Y = Math.Min(0.0, mi_MinMax.md_MinY * AXIS_EXCESS);
                        i_Axis.mi_Points[1].mi_P3D.md_Y = Math.Max(0.0, mi_MinMax.md_MaxY * AXIS_EXCESS);
                        // ---------------------------
                        i_Axis.mi_Points[0].mi_P3D.md_X = Math.Min(0.0, mi_MinMax.md_MinX * AXIS_EXCESS); // Y axis at minimum X position
                        i_Axis.mi_Points[1].mi_P3D.md_X = Math.Min(0.0, mi_MinMax.md_MinX * AXIS_EXCESS); // Y axis at minimum X position
                        i_Axis.me_Line   = eCoord.Y;
                        i_Axis.me_Offset = eCoord.Z; // Show zero label at Y axis
                        break;
                    case eCoord.Z: // Red
                        i_Axis.mi_Points[0].mi_P3D.md_Z = Math.Min(0.0, mi_MinMax.md_MinZ * AXIS_EXCESS);
                        i_Axis.mi_Points[1].mi_P3D.md_Z = Math.Max(0.0, mi_MinMax.md_MaxZ * AXIS_EXCESS);
                        // ---------------------------
                        i_Axis.mi_Points[0].mi_P3D.md_X = Math.Min(0.0, mi_MinMax.md_MinX * AXIS_EXCESS); // Z axis start at minimum X position
                        i_Axis.mi_Points[1].mi_P3D.md_X = Math.Min(0.0, mi_MinMax.md_MinX * AXIS_EXCESS); // Z axis start at minimum X position
                        i_Axis.mi_Points[0].mi_P3D.md_Y = Math.Min(0.0, mi_MinMax.md_MinY * AXIS_EXCESS); // Z axis start at minimum Y position
                        i_Axis.mi_Points[1].mi_P3D.md_Y = Math.Min(0.0, mi_MinMax.md_MinY * AXIS_EXCESS); // Z axis start at minimum Y position
                        i_Axis.me_Line   = eCoord.Z;
                        i_Axis.me_Offset = eCoord.Z; // Hide zero label at Z axis (Z,Z invalid)
                        break;
                }

                i_Axis.mi_Points[0].Project3D(this);
                i_Axis.mi_Points[1].Project3D(this);
                i_Axis.CalcAngle2D();

                i_Lines.Add(i_Axis);
            }

            // Calculate currently visible quadrant
            mi_Quadrant.Calculate(mi_Mouse.md_Phi, i_Lines[(int)eCoord.X], i_Lines[(int)eCoord.Y], i_Lines[(int)eCoord.Z]);

            // Add raster lines in 6 different directions
            if (me_Raster >= eRaster.Raster)
            {
                for (int A=0; A<3; A++) // iterate axis X,Y,Z
                {
                    // Combine X+Y, Y+Z, Z+X
                    eCoord e_First  = (eCoord)(A);
                    eCoord e_Second = (eCoord)((A+1) % 3);

                    for (int D=0; D<2; D++) // iterate first, second
                    {
                        cLine i_FirstAxis  = i_Lines[(int)e_First];
                        cLine i_SecondAxis = i_Lines[(int)e_Second];

                        double d_SecndStart = i_SecondAxis.mi_Points[0].mi_P3D.GetValue(e_Second);
                        double d_SecndEnd   = i_SecondAxis.mi_Points[1].mi_P3D.GetValue(e_Second);

                        // Distance between raster lines
                        double d_Interval = CalculateInterval(d_SecndEnd - d_SecndStart);

                        for (int L=-11; L<11; L++) // iterate raster line
                        {
                            double d_Offset = d_Interval * L;

                            if (d_Offset < d_SecndStart || d_Offset > d_SecndEnd) 
                                continue;
                                
                            cLine i_Raster = new cLine();
                            i_Raster.mi_Pen    = mi_Axis[(int)e_Second].mi_RasterPen;
                            i_Raster.me_Line   = e_First;
                            i_Raster.me_Offset = e_Second;
                            i_Raster.ms_Label  = FormatDouble(d_Offset);
                            i_Raster.mi_Points[0].mi_P3D = i_FirstAxis.mi_Points[0].mi_P3D.Clone();
                            i_Raster.mi_Points[1].mi_P3D = i_FirstAxis.mi_Points[1].mi_P3D.Clone();

                            i_Raster.mi_Points[0].mi_P3D.SetValue(e_Second, d_Offset);
                            i_Raster.mi_Points[1].mi_P3D.SetValue(e_Second, d_Offset);

                            // Do not draw a raster line which equals a main axis
                            if (i_Raster.CoordEquals(i_Lines[(int)eCoord.X]) ||
                                i_Raster.CoordEquals(i_Lines[(int)eCoord.Y]) ||
                                i_Raster.CoordEquals(i_Lines[(int)eCoord.Z]))
                                continue;

                            if ((e_First == eCoord.X && e_Second == eCoord.Z) || // Blue
                                (e_First == eCoord.Z && e_Second == eCoord.X))
                            {
                                i_Raster.md_Sort = mi_Quadrant.md_SortXZ;
                            }
                            else if ((e_First == eCoord.Z && e_Second == eCoord.Y) || // Green
                                     (e_First == eCoord.Y && e_Second == eCoord.Z))
                            {
                                i_Raster.md_Sort = mi_Quadrant.md_SortYZ;
                            }
                            else // X + Y Red
                            {
                                i_Raster.md_Sort = mi_Quadrant.md_SortXY;

                                // Special case: XY raster lines must be shifted down to negative end of Z axis
                                cLine i_AxisZ = i_Lines[(int)eCoord.Z];
                                i_Raster.mi_Points[0].mi_P3D.md_Z = i_AxisZ.mi_Points[0].mi_P3D.md_Z;
                                i_Raster.mi_Points[1].mi_P3D.md_Z = i_AxisZ.mi_Points[0].mi_P3D.md_Z;
                            }

                            i_Lines.Add(i_Raster);
                        }

                        // Swap first and second
                        eCoord e_Temp = e_First;
                        e_First  = e_Second;
                        e_Second = e_Temp;
                    }
                }
            }

            // Convert axis and raster lines 3D --> 2D
            foreach (cLine i_Line in i_Lines)
            {
                i_Line.mi_Points[0].Project3D(this);
                i_Line.mi_Points[1].Project3D(this);
                InsertDrawObject(i_Line);
            }

            // Move the graph to the left when labels are enabled
            if (me_Raster == eRaster.Labels)
            {
                int s32_LabelWidth = 0;
                foreach (cLine i_Line in i_Lines)
                {
                    if (i_Line.me_Line == eCoord.Y && i_Line.me_Offset == eCoord.Z)
                    {
                        SizeF  k_Size  = i_Graph.MeasureString(i_Line.ms_Label, Font);
                        s32_LabelWidth = Math.Max(s32_LabelWidth, (int)k_Size.Width);
                    }
                }
                mi_Mouse.mk_OffCoord.X -= s32_LabelWidth / 2;
            }
        }

        /// <summary>
        /// polygons, scatter points and axis must be sorted for correct drawing order!
        /// </summary>
        private void InsertDrawObject(cDrawObj i_DrawObj)
        {
            int P;
            for (P=0; P<mi_DrawObjects.Count; P++)
            {
                if (mi_DrawObjects[P].md_Sort > i_DrawObj.md_Sort)
                    break;
            }
            mi_DrawObjects.Insert(P, i_DrawObj);
        }

        /// <summary>
        /// Makes a color brigther
        /// </summary>
        private static Color BrightenColor(Color c_Color)
        {
            int s32_Red   = c_Color.R + (255 - c_Color.R) / 2;
            int s32_Green = c_Color.G + (255 - c_Color.G) / 2;
            int s32_Blue  = c_Color.B + (255 - c_Color.B) / 2;

            return Color.FromArgb(255, s32_Red, s32_Green, s32_Blue);
        }

        /// <summary>
        /// returns intervals of  0.1, 0.2, 0.5,  1, 2, 5,  10, 20, 50,  etc...
        /// The count of intervals which fit into the range is always between 5 and 10
        /// </summary>
        private static double CalculateInterval(double d_Range)
        {
            double d_Factor = Math.Pow(10.0, Math.Floor(Math.Log10(d_Range)));
            if (d_Range / d_Factor >= 5.0)
                return d_Factor;
            else if (d_Range / (d_Factor / 2.0) >= 5.0)
                return d_Factor / 2.0;
            else
                return d_Factor / 5.0;
        }

        // md_Label = 123.000 --> display "123"
        // md_Label =  15.700 --> display "15.7"  
        // md_Label =   4.260 --> display "4.26"
        // md_Label =   0.834 --> display "0.834"
        public static String FormatDouble(double d_Label)
        {
            return d_Label.ToString("0.000", CultureInfo.InvariantCulture).TrimEnd('0').TrimEnd('.');
        }

        // =================================== DRAWING =====================================

        protected override void OnPaintBackground(PaintEventArgs e)
        {
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            Render(e.Graphics);

            if (mi_BorderPen != null)
            {
                e.Graphics.ResetTransform();
                Rectangle r_Border = ClientRectangle;
                e.Graphics.DrawRectangle(mi_BorderPen, r_Border.X, r_Border.Y, r_Border.Width - 1, r_Border.Height - 1);
            }
        }

        public Bitmap GetScreenshot()
        {
            Bitmap i_Bmp = new Bitmap(ClientSize.Width, ClientSize.Height);
            using (Graphics i_Graph = Graphics.FromImage(i_Bmp))
            {
                Render(i_Graph);
            }
            return i_Bmp;
        }

        private void Render(Graphics i_Graph)
        {
            i_Graph.Clear(BackColor);

            if (mb_Updating)
                return;

            foreach (cMessgData i_Mesg in mi_MessageData)
            {
                i_Mesg.Draw(i_Graph, ClientRectangle);
            }

            if (mi_DrawObjects.Count == 0)
            {
                CreateCoordinateSystem(i_Graph);

                foreach (cRenderData i_Data in mi_RenderData) 
                {
                    i_Data.ProjectDrawObjects();
                }

                if (mi_DrawObjects.Count == 0) // no 3D data loaded
                    return;
            }

            // Draw axis legends at bottom
            int X = 4;
            int Y = ClientSize.Height - Font.Height - 4;
            for (int i=2; i>=0; i--)
            {
                if (String.IsNullOrEmpty(mi_Axis[i].ms_LegendText))
                    continue;

                String s_Disp = String.Format("{0}: {1}", (eCoord)i, mi_Axis[i].ms_LegendText);
                i_Graph.DrawString(s_Disp, Font, mi_Axis[i].mi_LegendBrush, X,  Y);
                Y -= Font.Height;
            }

            // Draw rotation legend at top
            if (mi_TopLegendBrush != null)
            {
                String[] s_Legend = new String[] { "Rotation:", "Elevation:", "Distance:" };
                String[] s_Value  = new String[] { String.Format("{0:+#;-#;0}°", (int)mi_Mouse.md_Phi),
                                                   String.Format("{0:+#;-#;0}°", (int)mi_Mouse.md_Theta),
                                                   String.Format("{0}",          (int)mi_Mouse.md_Rho) };

                SizeF k_Size = i_Graph.MeasureString(s_Legend[1], Font); // measure the widest string
                X = 4;
                Y = 3;
                for (int i=0; i<3; i++)
                {
                    i_Graph.DrawString(s_Legend[i], Font, mi_TopLegendBrush, X,  Y);
                    i_Graph.DrawString(s_Value [i], Font, mi_TopLegendBrush, X + k_Size.Width, Y);
                    Y += Font.Height;
                }
            }

            // Set X, Y offset which user has set by mouse dragging with SHIFT key pressed
            i_Graph.TranslateTransform(mi_Mouse.mk_OffMove.X + mi_Mouse.mk_OffCoord.X, 
                                       mi_Mouse.mk_OffMove.Y + mi_Mouse.mk_OffCoord.Y);

            SmoothingMode e_Smooth = SmoothingMode.Invalid;

            foreach (cDrawObj i_DrawObj in mi_DrawObjects)
            {
                if (!i_DrawObj.IsValid)
                    continue; // avoid overflow exception or hanging

                if (e_Smooth != i_DrawObj.me_SmoothMode) // avoid unneccessary calls into GDI+ (speed optimization)
                {
                    e_Smooth              = i_DrawObj.me_SmoothMode;
                    i_Graph.SmoothingMode = i_DrawObj.me_SmoothMode;
                }

                cLine i_Line = i_DrawObj as cLine;
                if (i_Line != null)
                {
                    i_Graph.DrawLine(i_Line.mi_Pen, i_Line.mi_Points[0].mi_P2D.Coord, i_Line.mi_Points[1].mi_P2D.Coord);

                    // ------------ Label ------------

                    if (i_Line.ms_Label != null            &&
                        me_Raster == eRaster.Labels        && 
                        mi_Quadrant.mb_BottomView == false && // no label in bottom view
                        mi_Quadrant.ms32_Quadrant == 3)       // only in quadrant 3 showing labels makes sense
                    {
                        PointF k_Pos = i_Line.mi_Points[1].mi_P2D.Coord;
                        StringFormat i_Align = new StringFormat();
                        if (i_Line.me_Line == eCoord.Y && i_Line.me_Offset == eCoord.Z)
                        {
                            k_Pos.X += 5;
                            k_Pos.Y -= Font.Height / 2;
                        }
                        else if (i_Line.me_Line == eCoord.Y && i_Line.me_Offset == eCoord.X)
                        {
                            k_Pos.X += (float)mi_Transform.ProjectXY(5, -5);
                            k_Pos.Y += (float)mi_Transform.ProjectXY(-Font.Height / 2, 5);
                        }
                        else if (i_Line.me_Line == eCoord.X && i_Line.me_Offset == eCoord.Y)
                        {
                            k_Pos.X += (float)mi_Transform.ProjectXY(5, -5);
                            k_Pos.Y += (float)mi_Transform.ProjectXY(5, -Font.Height / 2);
                            i_Align.Alignment = StringAlignment.Far;
                        }
                        else continue;

                        Brush  i_Brush = mi_Axis[(int)i_Line.me_Offset].mi_LegendBrush;
                        i_Graph.DrawString(i_Line.ms_Label, Font, i_Brush, k_Pos, i_Align);
                    }
                }
                else // cPolygon, cScatter
                {
                    i_DrawObj.Render(i_Graph);
                }
            } // foreach (cDrawObj)
        }

        // ============================== MOUSE =====================================

        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);
            mi_Tooltip.Hide();

            mi_Mouse.mk_LastPos = e.Location;

            if (mi_DrawObjects.Count == 0)
                return;
           
            switch (Control.ModifierKeys)
            {
                case Keys.None:
                    if (e.Button == MouseButtons.Left)
                    {
                        Cursor = Cursors.NoMoveVert;
                        mi_Mouse.me_Action = eMouseAction.Theta;
                    }

                    if (e.Button == MouseButtons.Right)
                    {
                        Cursor = Cursors.NoMoveHoriz;
                        mi_Mouse.me_Action = eMouseAction.Phi;
                    }
                    break;

                case Keys.Shift:
                    if (e.Button == MouseButtons.Left)
                    {
                        Cursor = Cursors.NoMove2D;
                        mi_Mouse.me_Action = eMouseAction.Move;
                    }
                    break;

                case Keys.Control:
                    if (e.Button == MouseButtons.Left)
                    {
                        Cursor = Cursors.SizeNS;
                        mi_Mouse.me_Action = eMouseAction.Rho;
                    }
                    break;
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);
            mi_Tooltip.Hide();

            mi_Mouse.me_Action = eMouseAction.None;
            Cursor = Cursors.Arrow;
        }

        protected override void OnMouseLeave(EventArgs e)
        {
            base.OnMouseLeave(e);
            mi_Tooltip.Hide();

            mi_Mouse.me_Action = eMouseAction.None;
            Cursor = Cursors.Arrow;
        }

        protected override void OnMouseWheel(MouseEventArgs e)
        {
            base.OnMouseWheel(e);
            mi_Tooltip.Hide();

            if (mi_Mouse.OnMouseWheel(e.Delta))
            {
                mi_Transform.SetCoefficients(mi_Mouse);

                mi_DrawObjects.Clear(); // recalculate
                Invalidate();           // repaint
            }
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            int s32_DiffX = e.X - mi_Mouse.mk_LastPos.X;
            int s32_DiffY = e.Y - mi_Mouse.mk_LastPos.Y;
            mi_Mouse.mk_LastPos = e.Location;

            switch (mi_Mouse.me_Action)
            {
                case eMouseAction.Move:
                    mi_Tooltip.Hide();
                    mi_Mouse.mk_OffMove.X += s32_DiffX;
                    mi_Mouse.mk_OffMove.Y += s32_DiffY;

                    Invalidate(); // repaint only
                    break;

                case eMouseAction.Rho:
                case eMouseAction.Theta:
                case eMouseAction.Phi:
                    mi_Tooltip.Hide();
                    mi_Mouse.OnMouseMove(s32_DiffX, s32_DiffY);
                    mi_Transform.SetCoefficients(mi_Mouse);

                    mi_DrawObjects.Clear(); // recalculate
                    Invalidate();           // repaint
                    break;

                case eMouseAction.None:
                    mi_Tooltip.OnMouseMove(e);
                    break;
            }
        }

        // ============================================================================

        protected override void OnCreateControl()
        {
            base.OnCreateControl();

            // This is the size of the control defined in Visual Studio Form Designer
            mi_Transform.SetInitialSize(ClientSize);

            if (BorderStyle != BorderStyle.None && BorderColor.A > 0)
            {
                MessageBox.Show(this, "Invalid Configuration:\n"
                              + "Either set a BorderStyle != None and the framework draws a gray border "
                              + "or set a BorderColor != Transparent and the control draws a border in your color.\n"
                              + "If you set both, two borders will be drawn.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);

            // This may be called with an invalid size before the control is created!
            mi_Transform.SetSize(ClientSize);

            mi_DrawObjects.Clear(); // recalculate
            Invalidate();           // repaint
        }

        /// <summary>
        /// This is only called when the user moves the trackbar, not when TrackBar.Value is set programmatically.
        /// </summary>
        void OnTrackbarScroll(object sender, EventArgs e)
        {
            mi_Mouse.OnTrackBarScroll();
            mi_Transform.SetCoefficients(mi_Mouse);

            mi_DrawObjects.Clear(); // recalculate
            Invalidate();           // repaint
        }
    }
}